/* 
	PsyQ Overlay Example
	Basic demonstration of PsyQ overlay functionality for use with 
	reverse-engineering tools.
	Copyright 2019 Ben Lincoln
	https://www.beneaththewaves.net/

	This file is part of PsyQ Overlay Example.

	PsyQ Overlay Example is free software: you can redistribute it and/or modify
	it under the terms of version 3 of the GNU General Public License as published by
	the Free Software Foundation.

	PsyQ Overlay Example is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with PsyQ Overlay Example (in the file LICENSE.txt).  
	If not, see <http://www.gnu.org/licenses/>.

   */
   
#include <libsn.h>

// base address for loading overlays
// defined in overlay.s
extern char *OverlayAddress;

// example functions in loadable overlays
extern void overlay1_function_1 (void);
extern void overlay1_function_2 (void);
extern void overlay2_function_1 (char *);
extern void overlay2_function_2 (void);
extern void overlay3_function_1 (char *);
extern void overlay3_function_2 (void);

// just an example variable to use in various places
int GlobalNumber;

// some functions for the loadable overlays to call
void PrintMessage(char *message)
{
	printf("%s\n", message);
}

void PrintCurrentGlobalNumber()
{
	printf("Current global number value is %i\n", GlobalNumber);
}

static void loadOverlay(char *fileName)
{
	int fileHandle;
	int fileLength;

	fileHandle = PCopen(fileName, 0, 0);
	fileLength= PClseek(fileHandle, 0, 2);
	PClseek(fileHandle, 0, 0);
	PCread(fileHandle, OverlayAddress, fileLength);
	PCclose(fileHandle);
	FlushCache();
}

int main()
{
	PrintMessage("PsyQ Overlay Example");
	GlobalNumber = 0;
	
	PrintMessage("Loading OVERLAY1.BIN");
	loadOverlay("OVERLAY1.BIN");
	PrintMessage("Calling overlay1_function_1()");
	overlay1_function_1();
	PrintMessage("Calling overlay1_function_2()");
	overlay1_function_2();
	
	PrintMessage("Loading OVERLAY2.BIN");
	loadOverlay("OVERLAY2.BIN");
	PrintMessage("Calling overlay2_function_1(\"Sent to overlay 2\")");
	overlay2_function_1("Sent to overlay 2");
	PrintMessage("Calling overlay2_function_2()");
	overlay2_function_2();
	
	PrintMessage("Loading OVERLAY3.BIN");
	loadOverlay("OVERLAY3.BIN");
	PrintMessage("Calling overlay3_function_1(\"Sent to overlay 3\")");
	overlay3_function_1("Sent to overlay 3");
	PrintMessage("Calling overlay3_function_2()");
	overlay3_function_2();
	
	PrintMessage("Trying to access an overlay which is no longer loaded should cause unexpected behaviour.");
	
	PrintMessage("Calling overlay1_function_1() even though overlay 1 has been overwritten with overlay 3");
	overlay1_function_1();
	PrintMessage("Calling overlay1_function_2() even though overlay 1 has been overwritten with overlay 3");
	overlay1_function_2();
	PrintMessage("Calling overlay2_function_1(\"Sent to overlay 2\") even though overlay 2 has been overwritten with overlay 3");
	overlay2_function_1("Sent to overlay 2");
	PrintMessage("Calling overlay2_function_2() even though overlay 2 has been overwritten with overlay 3");
	overlay2_function_2();
	
	PrintMessage("Done!");
	
	return 0;
}